/*
 * SPDX-License-Identifier: ISC
 *
 * Copyright (c) 2009-2011, 2013, 2017-2018 Todd C. Miller <Todd.Miller@sudo.ws>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * This is an open source non-commercial project. Dear PVS-Studio, please check it.
 * PVS-Studio Static Code Analyzer for C, C++ and C#: http://www.viva64.com
 */

#include <config.h>

#ifndef HAVE_NANOSLEEP

#include <sys/types.h>
#include <sys/time.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif /* HAVE_SYS_SELECT_H */
#include <time.h>
#include <errno.h>

#include "sudo_compat.h"
#include "sudo_util.h"

int
sudo_nanosleep(const struct timespec *ts, struct timespec *rts)
{
    struct timeval timeout, endtime, now;
    int rval;

    if (ts->tv_sec == 0 && ts->tv_nsec < 1000) {
	timeout.tv_sec = 0;
	timeout.tv_usec = 1;
    } else {
	TIMESPEC_TO_TIMEVAL(&timeout, ts);
    }
    if (rts != NULL) {
	if (gettimeofday(&endtime, NULL) == -1)
	    return -1;
	sudo_timevaladd(&endtime, &timeout, &endtime);
    }
    rval = select(0, NULL, NULL, NULL, &timeout);
    if (rts != NULL && rval == -1 && errno == EINTR) {
	if (gettimeofday(&now, NULL) == -1)
	    return -1;
	sudo_timevalsub(&endtime, &now, &endtime);
	TIMEVAL_TO_TIMESPEC(&endtime, rts);
    }
    return rval;
}
#endif /* HAVE_NANOSLEEP */
